<?php
  // $Id: civinode_utils.inc,v 1.30 2006/08/25 23:59:13 torenware Exp $
  /**
   *
   * CiviNode Utility Functions
   *
   * These include wrappers for common CRM related tasks like
   * determining group membership, going from NID to CID and
   * back, and doing loads.
   *
   */

  //lowlevel routines to bypass the normal API.
  //all of these routines should be treated as private
  //with extreme prejudice.
require_once 'lowlevel/civinode_groups.inc';

//Some magic numbers that Lobo won't code compatibly for us (rant rant)
define('CIVINODE_MAGIC_ACTION_NONE',                0);
define('CIVINODE_MAGIC_ACTION_ADD',                 1);
define('CIVINODE_MAGIC_ACTION_UPDATE',              2);
define('CIVINODE_MAGIC_ACTION_VIEW',                4);
define('CIVINODE_MAGIC_ACTION_DELETE',              8);
define('CIVINODE_MAGIC_ACTION_BROWSE',             16);
define('CIVINODE_MAGIC_ACTION_ENABLE',             32);
define('CIVINODE_MAGIC_ACTION_DISABLE',            64);
define('CIVINODE_MAGIC_ACTION_EXPORT',            128);
define('CIVINODE_MAGIC_ACTION_BASIC',             256);
define('CIVINODE_MAGIC_ACTION_ADVANCED',          512);
define('CIVINODE_MAGIC_ACTION_PREVIEW',          1024);
define('CIVINODE_MAGIC_ACTION_FOLLOWUP',         2048);
define('CIVINODE_MAGIC_ACTION_MAP',              4096);
define('CIVINODE_MAGIC_ACTION_PROFILE',          8192);
define('CIVINODE_MAGIC_ACTION_MAX_ACTION',      16383);



/**
 * Initialization checks
 *
 *
 */
function civinode_check_init(){
  if (function_exists('civicrm_initialize')) {
    civicrm_initialize(TRUE);
    return TRUE;
  }
  return FALSE;
}



/**
 * Check for admin level access for this operation
 */
function civinode_check_admin_access($op = 'view'){
  global $user;
  //user_access is heavily cached, so we don't
  //worry about doing any more of it
  if ( user_access('administer CiviCRM', $user))
    return TRUE;
  if ( $op == 'view' and user_access('view all contacts'))
    return TRUE;
  if ( ($op == 'update' or $op == 'delete') and  user_access('edit all contacts'))
    return TRUE;
    
  return FALSE;

}


/**
 * Error handling. We will want to keep error info
 * around so we can pass it to UI, etc. The setter
 * will do this.
 *
 * @defgroup Error_Handlers
 */

/**
 * Error testing and logging routine.
 *
 * @ingroup Error_Handlers
 * @param mixed $obj value returned by CiviCRM API that may be error.
 * @return boolean TRUE if $obj is a CiviCRM error.
 */
function civinode_check_error($obj = NULL){
  //Reset the error
  civinode_clear_error();
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }
  if (!$obj) {
    $obj = _civinode_get_last_error();
  }
  //Docs are wrong... :-(
  //if(crm_is_error($obj)){
  if($obj and is_a($obj, 'CRM_Core_Error')){
    $msg = $obj->_errors[0]['message'];
    //may want to record the calling URL as well...
    watchdog('CRM', $msg, WATCHDOG_ERROR);
    _civinode_set_error($msg);
    return TRUE;
  }
  else
    return FALSE;

}

/**
 * Get the text of the last error checked via civinode_check_error
 * 
 * @ingroup Error_Handlers
 * @return string Error string.
 */
function civinode_get_last_error(){
  return _civinode_set_error();
}

/**
 * Clear the last error saved.
 *
 * @ingroup Error_Handlers
 * 
 */
function civinode_clear_error(){
  _civinode_set_error(NULL, TRUE);
}


function _civinode_set_error($value = NULL, $reset = FALSE){
  static $error;
  $last_error = $error;
  if($reset){
    $error = NULL;
    return NULL;
  }
  if(!$value)
    return $last_error;
  else{
    $error = $value;
    return $value;
  }

}


/**
 * Non destructively see if you can get the last error
 * off the error stack, if there is one.
 * 
 * @ingroup Error_Handlers
 * @return  for now, a PEAR error object
 */
function &_civinode_get_last_error(){
  $last_error = NULL;
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return "FATAL: CiviCRM not installed";
  }

  $error_stack =& CRM_Core_Error::singleton();
  //For now, let's see how "deep the rabbit hole goes" and
  //return just the tip of the stack.
  if ($error_stack->hasErrors()) {
    $errors = $error_stack->getErrors();
    $lastError = array_pop($errors);
  }
  
  return $last_error;
}


/**
 * Metadata about contacts
 * 
 * @defgroup CiviCRM_MetaData
 */


/**
 * Get a list of profile names.
 *
 * @ingroup CiviCRM_MetaData
 * @param array Search criteria for profiles (UNUSED).
 * @return array of int => string
 */
function civinode_get_profiles($params = NULL){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }
  //Note: as of 1.4, we ignore params
  //We may want to use the auth table to reduce this
  //list (i.e., you don't have the right, you can't
  //assign it.
  $profiles = crm_uf_get_profile_groups();
  //winnow it down here later...
  return $profiles;
}

/**
 * Get the human-readable name of a profile.
 *
 * @ingroup CiviCRM_MetaData
 * @param int CiviCRM Profile ID
 * @return string
 */
function civinode_get_profile_title($pid){
  static $cache;
  if(isset($cache[$pid]))
    return $cache[$pid];
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $cache[$pid] = crm_uf_get_profile_title($pid);
  return $cache[$pid];
}



/**
 * Get the right profile ID for displaying this drupal user.
 *
 * @ingroup CiviCRM_MetaData
 * @param int Drupal UID
 * @return int CiviCRM Profile ID
 */
function civinode_get_default_profile_id($uid){
  $pid = variable_get('civinode_default_pid', 1);
  return $pid; //TO DO: make this settable for the individual
}


/**
 * Get the appropriate meta data for this PID and operation
 *
 * @ingroup CiviCRM_MetaData
 * @param int CiviCRM Profile ID
 * @param string $op Drupal-style operation name
 * @return array of field_name => array
 */
function civinode_get_profile_metadata($pid, $op = ''){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }
  static $cache;
  $action = CRM_CORE_ACTION_VIEW;
  switch($op){
  default:
    //We will translate Drupal ops into actions here
    $action = CRM_CORE_ACTION_VIEW;
    break;
  }
  //structure of cache: first by pid, then ACTION
  if(isset($cache[$pid])){
    if(isset($cache[$pid][$action]))
      return $cache[$pid][$action];
  }
  else
    $cache[$pid] = array();

  //what should visibility be set to??
  $raw_data = crm_uf_get_profile_fields($pid, FALSE, $action); 
  $title = crm_uf_get_profile_title($pid);
  $cache[$pid][$action] = $raw_data;
  
  return $cache[$pid][$action];
}


/**
 * This function is a nasty hack to get around the way location-related
 * fields get handled.  It returns a map from what names are used for 
 * fields in a node or the assoc array calls in the CRM API,
 * into the names used internally by CRM for formatting info.
 *
 * @ingroup CiviCRM_MetaData
 * @param int CiviCRM Profile ID
 * @param string $op Drupal-style operation name
 * @return array of string 
 */
function civinode_get_profile_field_list($profile_id, $op = 'view'){
  static $cache;
  $cache_key = "$op::$profile_id";
  if(isset($cache[$cache_key]))
    return $cache[$cache_key];
  $meta_data = civinode_get_profile_metadata($profile_id, $op);
  $fields = array();
  if($meta_data){
    $raw_fields = array_keys($meta_data);
    foreach($raw_fields as $raw_field){
      list($base, $loc_type_id) = explode('-', $raw_field);
      //does this preserve order??
      $fields[$base] = $raw_field;
    }
  }
  $cache[$cache_key] = $fields;
  return $fields;
}

/**
 * Get the appropriate meta data for this field for this PID and operation
 *
 * @ingroup CiviCRM_MetaData
 * @param int CiviCRM Profile ID
 * @param string $op Drupal-style operation name
 * @return array of string => mixed
 */
function civinode_get_field_info($pid, $fld_name, $op = 'view'){
  //This function is a candidate for yet more
  //caching, but see if we are spending lots of time here
  //first...

  $metadata = civinode_get_profile_metadata($pid, $op);
  //This would be easy, except that CRM plays games with locations.
  //Since it does, we will need to deal with multiple locations
  //and convert to the "base form"
  
  //First, maybe we won't have to deal with this problem
  if(isset($metadata[$fld_name]))
    return $metadata[$fld_name];
  //OK, not so lucky.  See if we are dealing with a location
  //dependent item
  $keys = array_keys($metadata);
  $real_key = "";
  foreach($keys as $key){
    list($base, $loc_type_id) = explode('-', $key);
    if($base == $fld_name){
      $real_key = $base;
      break;
    }
  }
  if($real_key)
    return $metadata[$real_key];

  return NULL; //we don't recognize this key
}


/**
 * Get a Drupal user's CRM contact_id.  Caching wrapper for crm_uf_get_match_id
 * 
 * @ingroup CiviCRM_MetaData
 * @param int Drupal UID
 * @return int CiviCRM Profile ID
 *
 */
function civinode_util_get_user_cid($uid){
  if(!$uid)
    return 0;
  static $cid_cache;
  if(!isset($cid_cache[$uid])){
    if (!civinode_check_init()){
      watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
      return FALSE;
    }
    $cid = crm_uf_get_match_id($uid);
    //No user should lack a cid, so we
    //may be registering.  Don't cache
    //to cover that case.
    if ($cid)
      $cid_cache[$uid] = $cid;
    else
      return 0;
  }
  return $cid_cache[$uid];
}


/**
 * Loads contact data using a profile as a filter
 *
 * @ingroup CiviCRM_MetaData
 * @param int $cid CRM Contact ID
 * @param int $pid CRM Profile ID
 * @return array assoc. array of contact data
 */
function civinode_util_load_cdata($cid, $pid = 0){
  //TO DO: This code is 1.4 dependent. If we back port,
  //we will need to add some search code
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $params = array('id' => $cid);
  //TO DO: add profile filter
  //$data = crm_fetch_contact($params);
  //Work around for CRM-822
  $contacts = crm_contact_search($params);
  //this is also counter to spec:
  if(!isset($contacts[0][$cid]))
    return array();
  $data = $contacts[0][$cid];
  //TO DO: we may need to map some of the items
  //to prevent clashes with Drupal
  return $data;
}


/**
 * Information about groups
 * @defgroup CiviCRM_Groups
 */


/**
 * List groups by various criteria
 *
 * @ingroup CiviCRM_Groups
 * @param array $params assoc array of search params.  NULL is all groups.
 * @return array of int => string
 */
function civinode_get_groups($params = NULL){
  //we just wrap crm_get_groups
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }
  $groups = crm_get_groups($params);
  if(civinode_check_error($group))
    return NULL;
  $group_list = array();
  foreach($groups as $gdata){
    $group_list[$gdata->id] = $gdata->title;
  }

  return $group_list;
  
}


/**
 * List groups for a contact.
 *
 * @ingroup CiviCRM_Groups
 * @param int $cid CiviCRM Contact ID
 * @return array of int => string
 */
function civinode_get_groups_for_cid($cid) {
  static $group_cache;
  if (!is_array($group_cache))
    $group_cache = array();
  $groups = array();
  if (!isset($group_cache[$cid])) {
    if (!civinode_check_init()){
      watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
      return FALSE;
    }
    $list = CiviNode_LL_getContactGroup($cid, 'Added');
    //$list = crm_contact_get_groups($cid, 'Added');
    foreach ($list as $eid => $data) {
      $groups[] = $data['group_id'];
    }
    $group_cache[$cid] = $groups;
  }
  else
    $groups = $group_cache[$cid];

  return $groups;
}


/**
 * Wrapper to add a contact to a group.
 *
 * @ingroup CiviCRM_Groups
 * @param int $group_id CiviCRM Group ID
 * @param int $cid CiviCRM Contact ID
 * @param string $source source to save to CRM data store.
 * @return boolean TRUE if the contact was successfully added.
 */
function civinode_add_contact_to_group($group_id, $contact_id, $source = 'civinode'){
  if (civinode_util_cid_in_group($contact_id, $group_id))
    return TRUE; //nothing to do.
  $group = civinode_get_group_by_id($group_id);
  if (!$group or civinode_check_error($group))
    return FALSE;
  $contact = crm_get_contact(array('contact_id' => $contact_id));
  if (!$contact or civinode_check_error($contact))
    return FALSE;
  //Work-around for a 1.4 bug: id is set here, but not contact_id
  if (!isset($contact->contact_id))
    $contact->contact_id = $contact->id;
  $contacts = array($contact);
  //'API' could be whatever you want, including the name of your module
  crm_add_group_contacts($group, $contacts, 'Added', 'API');
  return TRUE;
}


/**
 * Remove a contact from a group
 *
 * @ingroup CiviCRM_Groups
 * @param int $group_id CiviCRM Group ID
 * @param int $cid CiviCRM Contact ID
 * @return boolean TRUE if the contact was successfully removed (or was not in this group).
 */
function civinode_remove_contact_from_group($gid, $cid){
  $group = civinode_get_group_by_id($id);
  if (!$group)
    return FALSE;
  if (!civinode_util_cid_in_group($cid, $gid))
    return TRUE; //nothing to do, but we are also OK
  $to_remove = array();
  $cobj = crm_get_contact(array('contact_id' => $cid));
  if (civinode_check_error($cobj))
    return FALSE;
  $to_remove[] =& $cobj;
  $rslt = ccrm_delete_group_contacts($group, $to_remove, 'API');
  if (!$rslt)
    return TRUE; //success
  else
    return !civinode_check_error($rslt); //so we log the error  
    
}


/**
 * Get a group by name, creating it if need be
 *
 * @ingroup CiviCRM_Groups
 * @param string $name name of the group to retrieve or create. 
 * @param string $source what module created group (UNUSED?)
 * @param boolean $create if true and no group exists, create a group.
 * @return int The CiviCRM Group ID of the group (false/0 if failed)
 */
function civinode_get_group_by_name($name, $source = 'civinode',  $create = TRUE){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $groups = crm_get_groups(array('title' => $name));
  if(civinode_check_error($groups))
    return NULL;
  $gid = 0; //pessimism
  if (count($groups)) {
    //find an exact match
    foreach ($groups as $group) {
      if ($name == $group->title) {
        $gid = $group->id;
        break;
      }
    }

  }
  else if ($create) {
    //Need a fresh minted group
    $params = array('title' => $name,
                    'source' => $source,
                    'is_active' => 1);
    $group = crm_create_group($params);
    if(civinode_check_error($group))
      return NULL;
    $gid = $group->id;
  }

  return $gid;

}


/**
 * Wrapper to fetch a CRM group object
 *
 * @ingroup CiviCRM_Groups
 * @param int $gid CiviCRM Group ID
 * @return object the CiviCRM group object (NULL on error)
 */
function civinode_get_group_by_id($gid){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $groups = crm_get_groups(array('id' => $gid));
  if(civinode_check_error($groups))
    return NULL;
  //Is this a behavior change from 1.3???
  return count($groups) ? $groups[0] : NULL;
}



/**
 * Get target groups as allowed for the current user
 *
 * @ingroup CiviCRM_Groups
 * @return array of int => string.
 */
function civinode_get_drupal_allowed_groups() {
	//TO DO: this can be greatly sped up.
  $gids = civinode_get_current_target_groups('view');
  $groups = array();
  if (count($gids)) {
    foreach ($gids as $gid) {
      $group = civinode_get_group_by_id($gid);
      $groups[$gid] = $group->title;
    }
  }
  return $groups;
}


/**
 * Determines if a CRM group contains a contact.
 *
 * @ingroup CiviCRM_Groups
 * @param int $cid CRM Contact ID
 * @param int $gid CRM Group ID
 * @return bool TRUE if it belongs
 */
function civinode_util_cid_in_group($cid, $gid){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return FALSE;
  }

  //TO DO: use a search instead.

  //What groups does this cid know about?
  $list = CiviNode_LL_getContactGroup($cid, 'Added');
  //$list = crm_contact_get_groups($cid, 'Added');
  //AFAIR keys are gid
  $groups = array_keys($list);
  return in_array($gid, $groups);
}




/**
 * Wrapper for getting back an array of contact ids.
 *
 * @ingroup CiviCRM_Groups
 * @param int $gid CiviCRM group ID
 * @param int $start index of first record 
 * @param int $num_recs number of records to pull. 0 is "all records"
 *
 */
function civinode_util_group_contacts($gid, $start = 0, $num_recs = 0){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $group = crm_get_groups(array('id' => $gid));
  $list = array();
  //establish a limit, due to memory bugs
  //in 1.4.
  $limit = $num_recs ? $num_recs : 1000; 
  if(civinode_check_error($group))
    return NULL;
  if (!count($group))
    return array();
  $contacts = crm_get_group_contacts($group[0],
                                     //array('id' => 1), //bare min please
                                     NULL, //default instead
                                     'Added', NULL, //status and sort
                                     $start, $limit); //up to first 10K records
  foreach($contacts as $contact){
    $list[] = $contact->contact_id;
  }
  return $list;

}


/**
 * Authentication routines.  These apply to CiviNode nodes only as of
 * the 4.7 version.  This will likely be pulled out as a separate
 * module, since most applications are not using these features.
 *
 * @defgroup CiviNode_Node_Access 
 */

// TO DO: Should accesses really be nodes?? 


/**
 * List of access fields along with their UI names
 *
 * @ingroup CiviNode_Node_Access 
 * @return array of string => string 
 */
function civinode_get_access_field_names() {
  static $table;
  if (!$table)
    $table = array( 'nid' => t('Access ID'),
                    'requester_gid' => t('Access For Group'),
                    'target_gid' => t('To Target Group'),
                    'view_access' => t('View Group'),
                    'add_access' => t('Add To Group'),
                    'remove_access' => t('Remove From Group'),
                    'update_access' => t('Update Group Contacts'),
                    'title' => t('Access Type'),
                    'access_desc' => t('Description')
                    );
  return $table;
}



/**
 * Check if the specified user can do an operation on a contact.
 *
 * @ingroup CiviNode_Node_Access
 * @param object $user Drupal user object
 * @param int $cid Contact ID of the target contact
 * @param string $op Drupal-style operation we want to test.
 * @return TRUE if the operation is allowed on this contact.
 */
function civinode_user_has_privilege_on_contact($user, $cid, $op = 'view'){
  //First, cover admins. user_access caches pretty aggressively,
  //so we don't have to.
  if ( user_access('administer CiviCRM', $user))
    return TRUE;
  if ( $op == 'view' and user_access('view all contacts'))
    return TRUE;
  if ( ($op == 'update' or $op == 'delete') and  user_access('edit all contacts'))
    return TRUE;
  if (!$user->uid) 
    return FALSE;
  $user_cid = civinode_util_get_user_cid($user->uid);
  $groups = civinode_get_groups_for_cid($user_cid);
  $targ_groups = civinode_get_groups_for_cid($cid);
  //TO DO: use CRM's privs from admin/access
  //Also, make sure there is a group of all contacts defined in CRM
  if ( !count($groups) or !count($targ_groups))
    return FALSE;
  else {
    $sql = "SELECT count(*) FROM {civinode_access} ";
    $in_str = " requester_gid IN (" . implode(',', $groups) . ") ";
    $in_str .= "AND target_gid IN (" . implode(',', $targ_groups) . ") ";
    $where_str = " WHERE ";
    switch ($op) {
    case 'create':
      $where_str .= "add_access != 0 AND ";
    case 'view':
      $where_str .= "view_access != 0 AND ";
      break;
    case 'update':
      $where_str .= "update_access != 0 AND ";
      break;
    case 'delete':
      $where_str .= "remove_access != 0 AND ";
      break;
    default:
      //This would be "any access at all"
      break;
    }

    $sql .= $where_str . $in_str;
    $rslt = db_result(db_query($sql));
    //If we found one record that fits, we are golden
    return $rslt > 0;
  }
  //Otherwise, no luck
  return FALSE; //no can see
}


/**
 * Check if the specified user can do an operation on a group of contacts.
 *
 * @ingroup CiviNode_Node_Access
 * @param object $user Drupal user object
 * @param int $gid Group ID of the target contacts
 * @param string $op Drupal-style operation we want to test.
 * @return TRUE if the operation is allowed on this group of contacts.
 */
function civinode_user_has_privilege_on_group($user, $gid, $op = 'view'){
  //First, cover admins. user_access caches pretty aggressively,
  //so we don't have to.
  if ( civinode_check_admin_access($op))
    return TRUE;

  //Experimental change: I'm going to use the standard
  //node access table instead of my custom one

  $groups = civinode_get_groups_for_cid($user->contact_id);
  //TO DO: use CRM's privs from admin/access
  //Also, make sure there is a group of all contacts defined in CRM
  if ( !count($groups))
    return FALSE;
  else {
    $sql = "SELECT count(*) FROM {civinode_access} ";
    $in_str = " requester_gid IN (" . implode(',', $groups) . ") ";
    $in_str .= "AND target_gid = %d";
    $where_str = " WHERE ";
    switch ($op) {
    case 'view':
      $where_str .= "view_access != 0 AND ";
      break;
    case 'update':
      $where_str .= "add_access != 0 AND ";
      break;
    case 'delete':
      $where_str .= "remove_access != 0 AND ";
      break;
    default:
      //This would be "any access at all"
      break;
    }

    $sql .= $where_str . $in_str;
    $rslt = db_result(db_query($sql, $gid));
    //If we found one record that fits, we are golden
    return $rslt > 0;
  }
  //Otherwise, no luck
  return FALSE; //no can see
}



/**
 * Get a list of accesses.  You can search on any field in the
 * civinode_access table.
 *
 * @ingroup CiviNode_Node_Access
 * @param $params array of string => mixed
 * @return array of object
 */
function civinode_get_accesses($params = array()){
  $accesses = array();
  //we may have some bogus fields in
  //params, so get the real ones
  $real_params = array();
  //$where_clause = ; 
  //get everything by default
  $where_clause = ""; 
  if (count($params)) {
    $fields = civinode_get_access_field_names();
    $is_first = TRUE;
    foreach ($params as $key => $val) {
      if (isset($fields[$key]))
        $real_params[$key] = $val;
      if ($is_first) {
        $is_first = FALSE;
        $where_clause .= "$key = '%s' ";
      }
      else {
        $where_clause .= "AND $key = '%s' ";
      }
      
    }
    
  }
  //Treat everything as a string
  //TO DO: support wild cards
  $sql = "SELECT * FROM {civinode_access} " . $where_clause . " ORDER BY requester_gid";
  $rslt = db_query($sql, $real_params);
  if ($rslt) {
    while ($obj = db_fetch_object($rslt)) {
      $accesses[] = $obj;
    }
  }
  return $accesses;
}


/**
 * Get a list of accesses for the civinode realm for a user
 *
 * @ingroup CiviNode_Node_Access
 * @param int $uid Drupal UID to check
 * @return array of node IDs for the access objects.
 */
function civinode_get_accesses_for_user($uid){
  $accesses = array();
  //test, since we are ignoring uid...
  $groups = civinode_get_current_group_memberships();
  if (!count($groups))
    return $accesses;
  $in = implode(',', $groups);
  $sql = "SELECT DISTINCT(ca.nid) AS access_id FROM {civinode_access} ca, {civinode_group_memberships} cgm WHERE cgm.uid = %d AND ca.requester_gid = cgm.group_id";
  $rslt = db_query($sql, $uid);
  while ($row = db_fetch_array($rslt)) {
    $accesses[] = $row['access_id'];
  }
  return $accesses;
}

/**
 * Get group memberships from the membership table. If it
 * isn't the current user, get the groups from CRM instead. This is
 * used to enforce CiviCRM group accesses via the Drupal node access
 * system.
 *
 * @ingroup CiviNode_Node_Access
 * @return array of CiviCRM Group ID.
 */
function civinode_get_current_group_memberships(){
  global $user;
  if (!$user->uid)
    return array();
  $sql = "SELECT group_id FROM {civinode_group_memberships} WHERE uid = %d";
  $rslt = db_query($sql, $user->uid);
  $groups = array();
  while ($row = db_fetch_array($rslt)) {
    $groups[] = $row['group_id'];
  }
  return $groups;
}


/**
 * Get accessible target groups for the current user
 *
 * @ingroup CiviNode_Node_Access
 * @param string $op Access to test
 * @return array of CiviCRM Group ID.
 */
function civinode_get_current_target_groups($op = 'view'){
  if (civinode_check_admin_access($op)) {
    //If we have enough access, give us all of the groups
    $group_list = civinode_get_groups();
    return array_keys($group_list);
  }
    
  $mem_groups = civinode_get_current_group_memberships();
  if (!count($mem_groups))
    return array();
  $in = implode(',', $mem_groups);
  $sql = "SELECT target_gid FROM {civinode_access} WHERE requester_gid IN ($in) ";
  switch ($op) {
  case 'view':
    $perm = "AND (view_access != 0 OR add_access != 0 OR remove_access != 0)";
    break;
  case 'update':
    $perm = "AND (update_access != 0)";
    break;
  case 'delete':
    $perm = "AND (delete_access != 0)";
    break;
  default:
    $perm = "";
    break;
  }
  
  $rslt = db_query($sql . $perm);
  $groups = array();
  while ($row = db_fetch_array($rslt)) {
    $groups[] = $row['target_gid'];
  }
  return $groups;
}



/**
 * Routines to aid authentication.  Drupal uses a table called
 * node_access to control and speed up auth queries. To make it
 * possible to fit into this rather static system,  we maintain
 * a user look-up table that lets the node access system reference
 * our access scheme
 *
 * @ingroup CiviNode_Node_Access
 * @param int $uid Drupal User ID
 * @param array of CiviCRM Group ID
 */


function civinode_register_user_groups($uid, $groups = array()){
  //updates civinode_group_memberships
  $sql = "DELETE FROM {civinode_group_memberships} WHERE uid = %d";
  db_query($sql, $uid);
  $contact_id = civinode_util_get_user_cid($uid);
  if (count($groups) and $contact_id) {
    $sql = "INSERT INTO {civinode_group_memberships} (uid, contact_id, group_id) VALUES (%d, %d, %d) ";
    $first = TRUE;
    foreach ($groups as $gid) {
      db_query($sql, $uid, $contact_id, $gid);
    }
  }
}


/**
 * Creates a record in node_access that links mirrored
 * node with its civinode access information.
 * 
 * @ingroup CiviNode_Node_Access
 * @param int $access_id  the nid of the civinode access object
 * @param int $nid  the nid of civinode contact object
 */
function civinode_update_node_access($access_id, $nid){
  $access = node_load($access_id);
  if (!$access)
    return;
  $sql = "REPLACE INTO {node_access} (nid, gid, realm,  grant_view, grant_update, grant_delete) VALUES(%d,%d,'civinode',%d,%d,%d)";
  db_query($sql,$nid,$access_id,
           $access->view_access, 
           $access->update_access, 
           $access->remove_access);

}


/**
 * Creates node accesses once a new civinode access is defined, to keep
 * our access tables up-to-date.
 *
 * @ingroup CiviNode_Node_Access
 * @param object $access CiviNode access object.
 */
function civinode_create_node_accesses($access) {
  //Find any nodes that belong to our target group
  $nodes = array_values(civinode_get_node_memberships($access->target_gid));
  foreach ($nodes as $nid) {
    civinode_update_node_access($access->nid, $nid);
  }
}


/**
 * The access has been updated; reflect this in
 * the node_access table
 *
 * @ingroup CiviNode_Node_Access
 * @param object $access the node object for this access
 */
function civinode_reflect_access_change($access) {
  //an access has changed; make sure that this is
  //reflected in the node_access table
  $sql = "
UPDATE {node_access} SET 
grant_view = %d, 
grant_update = %d, 
grant_delete  = %d
WHERE realm = 'civinode' AND gid = %d
";
  db_query($sql,
           $access->view_access, 
           $access->update_access, 
           $access->remove_access,
           $access->nid);
}


/**
 * The access was deleted, so we need to remove
 * references to it from node_access
 *
 * @ingroup CiviNode_Node_Access
 * @param int $access_id node ID of the late lamented access object
 */
function civinode_remove_access($access_id){
  $sql = "DELETE FROM {node_access} WHERE gid = %d AND realm = 'civinode'";
  db_query($sql, $access_id);
}


/**
 * A CRM group has just been deleted, remove any linked access nodes.
 *
 * @ingroup CiviNode_Node_Access
 * @param int $group_id CiviCRM access group that just disappeared.
 */
function civinode_remove_group_accesses($group_id){
  $sql = "SELECT nid FROM {civinode_access} WHERE target_gid = %d";
  $rslt = db_query($sql, $group_id);
  while ($row = db_fetch_array($rslt)) {
    //delete the nodes; hook_delete
    //for the access type will clean up
    //the node_access table
    node_delete($row['nid']);
  }
}



/**
 * Make sure that a mirrored contact has the needed
 * records in node_access
 *
 * @ingroup CiviNode_Node_Access
 * @param object $node a node of type 'civinode_contact'
 */
function civinode_update_node_accesses($node){
  if ($node->type != 'civinode_contact')
    return;
  $cid = $node->contact_id;
  if (!$cid)
    return;
  $groups = civinode_get_groups_for_cid($cid);
  //Now clear node_access of any preexisting grants
  $sql = "DELETE FROM {node_access} WHERE realm = 'civinode' AND nid = %d";
  db_query($sql, $node->nid);
  //and add them back.  
  if (!count($groups))
    return; //no groups, no accesses
  //Now get all of our accesses
  $in_groups = implode(',', $groups);
  $sql = "SELECT nid FROM {civinode_access} WHERE target_gid IN ($in_groups)";
  $rslt = db_query($sql);
  //Add each access the node is known to have
  while ($row = db_fetch_array($rslt)) {
    civinode_update_node_access($row['nid'], $node->nid);
  }

}

/**
 * One of our nodes is going away; remove it from node_access
 *
 * @ingroup CiviNode_Node_Access
 */
function civinode_drop_node_accesses($nid){
  $sql = "DELETE FROM {node_access} WHERE realm = 'civinode' AND nid = %d";
  db_query($sql, $nid);
}



/**
 * A CiviCRM group has changed somehow. Recheck all dependent
 * accesses.
 *
 * @ingroup CiviNode_Node_Access
 * @param int $gid a CiviCRM group_id
 */
function civinode_update_node_accesses_for_group($gid){

  //First, what are the dependent accesses?
  $sql = "SELECT nid AS access_id FROM {civinode_access} WHERE target_gid = %d";
  $accesses = array();
  $rslt = db_query($sql, $gid);
  while ($row = db_fetch_array($rslt)) {
    $accesses[] = $row['access_id'];
  }
  if (!count($accesses))
    return; //nothing to update

  //What's in the list now?
  $lookup =  civinode_get_node_memberships($gid);
  //we need nid => cid, so
  $node_lookup = array_flip($lookup); 

  //Now, get a list of nodes currently using those accesses.
  //We delete the ones we no longer need, and restore the rest.
  $in_clause = implode(',', $accesses);
  $current_nodes = array();
  $find_sql = "SELECT na.nid, na.gid as access_id FROM {node_access} na " . 
    "WHERE na.realm = 'civinode' AND  na.gid IN ($in_clause)";
  $delete_sql = "DELETE FROM {node_access} WHERE realm = 'civinode' AND  nid = %d AND gid = %d";
  $rslt = db_query($find_sql);
  while ($row = db_fetch_array($rslt)) {
    if (!isset($node_lookup[$row['nid']])) {
      //node is no longer in group, recind
      //its access grants
      db_query($delete_sql, $row['nid'], $row['access_id']);
    }
    else
      $current_nodes[$row['nid']][] = $row['access_id'];
  }

  //We've deleted anything we don't need, now add new
  //accesses where we do.
  foreach ($node_lookup as $nid => $cid) {
    $new_accesses = isset($current_nodes[$nid]) ?
      array_diff($accesses, $current_nodes[$nid]) :
      $accesses;
    foreach ($new_accesses as $access_id)
      civinode_update_node_access($access_id, $nid);
  }
  
}


/**
 * Register a civinode contact into its member groups
 *
 * @ingroup CiviNode_Node_Access
 */
function civinode_register_node_groups($nid, $cid){
  //Set up the civinode_node_memberships table for this node
  $groups = civinode_get_groups_for_cid($cid);
  db_query('DELETE FROM {civinode_node_memberships} WHERE contact_id = %d', $cid);
  $sql = "INSERT INTO {civinode_node_memberships} (nid, contact_id, group_id) VALUES (%d,%d,%d)";
  foreach ($groups as $gid) {
    db_query($sql, $nid, $cid, $gid);
  }
}

/**
 * Our node is going away, remove it from node membership lists
 *
 * @ingroup CiviNode_Node_Access
 */
function civinode_delete_node_groups($nid){
  db_query('DELETE FROM {civinode_node_memberships} WHERE nid = %d', $nid);
}



/**
 * Generates a look-up table for contact_id => nid for
 * the group $gid from the civinode_node_membership table.
 *
 * @ingroup CiviNode_Node_Access
 */
function civinode_get_node_memberships($gid){
  $sql = "SELECT * FROM {civinode_node_memberships} WHERE group_id = %d";
  $rslt = db_query($sql, $gid);
  $list = array();
  while ($row = db_fetch_array($rslt)) {
    $list[$row['contact_id']] = $row['nid'];
  }
  return $list;
}


/**
 * If nodes have been added to a group in CiviCRM, update our internal table.
 *
 * @ingroup CiviNode_Node_Access
 */
function civinode_add_node_memberships($gid, $contact_list){
  //Are any nodes associated with this group?
  $lookup = civinode_get_node_memberships($gid);
  //Loop over the contact IDs and if they have associated
  //nodes, add them to this group's list.
  $sql = "INSERT INTO {civinode_node_memberships} (nid, contact_id, group_id) VALUES (%d,%d,%d)";
  foreach ($contact_list as $cid) {
    if ( !isset($lookup[$cid]) ) {
      //See if a node exists
      $node = civinode_util_node_factory($cid, FALSE);
      if (isset($node->nid))
        db_query($sql, $node->nid, $cid, $gid);
    }
  }
}


/**
 * If nodes have been removed to a group in CiviCRM, update our internal table.
 *
 * @ingroup CiviNode_Node_Access
 */
function civinode_remove_node_memberships($gid, $contact_list){
  $lookup = civinode_get_node_memberships($gid);
  $sql = "DELETE FROM {civinode_node_memberships} WHERE group_id = %d AND contact_id = %d";
  foreach ($contact_list as $cid) {
    if ( isset($lookup[$cid]) ) {
      db_query($sql, $gid, $cid);
    }
  }
}


/**
 * Activity wrappers
 * 
 * @defgroup CiviCRM_Activity_History
 */


/**
 * Creates a simple activity history record
 *
 * @ingroup CiviCRM_Activity_History
 * @param int $cid  CRM Contact ID to associate with the activity
 * @param int $id_to_save  An integer key you want to save with the activity
 * @param string $activity_type  'Meeting' or other defined activity type
 * @param string $module_base  UI name of the module storing this stuff.
 * @param string $desc  Short description to display about the activity
 * @param int $group_id  An optional CRM group_id to associate with the activity
 * @param string $call_back  The name of a PHP function with the following
 *               prototype, returning a URL:
 *               
 *               function my_details_handler($it_to_save, $crm_activity_hist_id);
 * 
 *               You can use civinode_get_activity($crm_activity_hist_id) to get back
 *               an associative array with the data stored in the activity record.
 *                 
 *
 * @param int $date  Unix time stamp.  Default is now, i.e., time().
 * @return boolean  TRUE on success
 */
function civinode_create_activity($cid, $id_to_save, $activity_type,
                                  $module_base = 'civinode',
                                  $desc = "", $group_id = 0,
                                  $call_back = NULL, $date = NULL){
  $params = array('module' => $module_base,
                  'entity_table' => 'civicrm_contact',
                  'entity_id' => $cid,
                  'activity_summary' => $desc,
                  'activity_type' => $activity_type);
  if (!$date)
    $date = time();

  $date_str = date('Ymd000000', $date);
  $params['activity_date'] = $date_str;
  $params['activity_id'] = $id_to_save;
  //callback...
  //$url = url($link);
  if ($call_back)
    $params['callback'] = $call_back;
  $hist = crm_create_activity_history($params);
  return civinode_check_error($hist);
}


/**
 * Getter for activity history data
 *
 * @ingroup CiviCRM_Activity_History
 */
function civinode_get_activity($crm_activity_hist_id){
  $defaults = array();
  $params = array('id' => $crm_activity_hist_id);
  $activity = crm_get_activity_history_object($params, $defaults);
  if (civinode_check_error($activity))
    return NULL;
  return $defaults;
}



/**
 * Sample URL callback for activities
 *
 * @ingroup CiviCRM_Activity_History
 */
function  civinode_activity_callback($saved_id, $crm_activity_id) {
  error_log("Call back was called");
  $defaults = civinode_get_activity($crm_activity_id);
  if (!civinode_check_error($activity)) {
    //We are cheating: we assume that the module is really
    //a path
    if (isset($defaults['module'])) {
      $base = $defaults['module'] . "/activity/$crm_activity_id/$saved_id";
      return url($base);
    }
    
  }
}


//TO DO:  we may want to do synchronization via cron as well.  PITA. :-(

/**
 * Routines to create a "CiviCRM Contact" node type
 *
 * @defgroup CiviNode_Nodifier
 */



/**
 * Load the Drupal node that corresponds to the contact. Creates
 * the node object by default. Note that we are not checking
 * access rights here.
 *
 * @ingroup CiviNode_Nodifier
 */
function civinode_util_node_factory($cid, $create = TRUE){
  global $user;
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  //First, see if we know this CID already
  $sql = "SELECT * FROM {civinode_lookup} WHERE contact_id = %d";
  $row = db_fetch_array(db_query($sql, $cid));
  $node_data = array();
  if(!isset($row['nid'])){
    if(!$create)
      return NULL;
    //TODO: Officially, these nodes are published by user 1
    $node_data = array('contact_id' => $cid, 'type' => 'civinode_contact',
                       'status' => 1, 'uid' => $user->uid);
    $node = NULL;
  }
  else{
    $node = node_load($row['nid']);
  }

  $data = civinode_util_load_cdata($cid);
  if(!$data){
    //our contact disappeared on us
    //TO DO: delete stale data in lookup table
    return NULL;
  }
  if($node_data){
    $node_data["title"] = $data["display_name"];
    $node = (object)$node_data;
    node_save($node);
    //Make sure that hook_load gets called
    $node = node_load($node->nid);
    /**
     * //I'm guessing this gets done in node_save
     $sql = "INSERT INTO {civinode_lookup} (nid, contact_id) " .
     "VALUES(%d, %d)";
     if(isset($node->nid)) //paranoia
     db_query($sql, $node->nid, $cid);
     }
  
     if($node){
     foreach($data as $key => $val){
     $node->$key = $val;
     }
    ***/

  }
  return $node;
  
}

/**
 * Search routine for supporting Drupal search on contacts
 *
 * @ingroup CiviNode_Nodifier
 */
function civinode_search_by_key($keys){
  global $user;
  $results = array();
  //Gotta be a key
  if(isset($keys[0])){
    //TO DO: we need to screen by group membership for auth
    if (!civinode_check_init()){
      watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
      return NULL;
    }
    //OR is really needed here. Lobo??
    $return_fields = array('display_name' => 1, 
                           'id' => 1,
                           'contact_type' => 1,
                           );
    //$sort = array('display_name' => 'ASC');
    $found = array(); //for doing a union
    //Apparently, $keys is a string
    $key_array = explode(' ', $keys);
    foreach($key_array as $key){
      if(!$key)
        continue; //extra spaces in $keys
      $params = array('sort_name' => "%$key%");
      $rslt = crm_contact_search($params, $return_fields);
      foreach($rslt[0] as $id => $data){
        if(isset($found[$id]))
          continue;
         
        //TO DO: check group membership here??
        if (!civinode_user_has_privilege_on_contact($user, $id, 'view'))
          continue;
        $s_rslt = array('link' => "/civinode/contact/$id",
                        'type' => $data['contact_type'],
                        'title' => $data['display_name']);
        $found[$id] = $s_rslt;
      }

    }
    //TO DO: sort + access
    $results = array_values($found);
  }
  return $results;
}

/**
 * Form API widgets and stuff
 *
 * @defgroup CiviNode_Form_Widgets
 */



/**
 * AJAX autocompleter handler for CiviCRM Contacts
 *
 * @ingroup CiviNode_Form_Widgets
 */
function civinode_contact_autocompleter($key, $val, $string){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), 
             WATCHDOG_ERROR);
    exit();
  }
  //first, require that we have at least 3 characters
  if (strlen($string) < 3)
    exit();

  //OR is really needed here. Lobo??
  $return_fields = array('display_name' => 1, 
                         'id' => 1,
                         'contact_type' => 1,
                         );
  
  $params['sort_name'] = "%$string%";
  //Check the key/val argument for special handling
  if ($key == 'gid' and is_numeric($val)) {
    $params['group'] = array($val => 1);

  }
  $rslt = crm_contact_search($params, $return_fields);
  $matches = array();
  foreach($rslt[0] as $id => $data){
    $id_str = " (" . $id . ")";
    $matches[$data['display_name'] . $id_str] = check_plain($data['display_name'] . $id_str);
  }
  //This assumes you are using
  //4.7 RC2 or later...
  print drupal_to_js($matches);
  exit();

}



/**
 * Returns a keyed array of id => 'useful display name' for all basic CRM types
 * It assumes that we want to check for access permissions for the global
 * Drupal user (for contacts and groups). This function is useful for presenting
 * selection options to users during input forms, etc.
 *
 * @ingroup CiviNode_Form_Widgets
 */
function civinode_get_allowed_options($type, $params = NULL) {
  static $cache;
  if(isset($cache[$type]) and !$params) {
    return $cache[$type];
  }
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }
  $result = array();
  switch($type) {
  case 'relationship':
    $relationships = crm_get_relationship_types();
    while($relationship = array_pop($relationships)) {
      $result[$relationship->id] = $relationship->description;
    }
    break;
    
  case 'group':
    $result = civinode_get_drupal_allowed_groups();
    break;
      
  case 'profile':
    $result = crm_uf_get_profile_groups();
    break;
    
  case 'contact':
    $allowed_groups = civinode_get_drupal_allowed_groups();
    $groups = array();
    //we restrict to the requested groups if there are any.
    //0 is the "universal" group, so skip this if it is requested
    if ($params and isset($params['groups']) and !in_array(0, $params['groups'])) {
      foreach ($params['groups'] as $requested_group) {
        if (isset($allowed_groups[$requested_group]))
          $groups[$requested_group] = $allowed_groups[$requested_group];
      }
    }
    else
      $groups = $allowed_groups;
    foreach($groups as $gid => $title) {
       
      $tmp = crm_get_groups(array('id' => $gid));
      $group = array_pop($tmp);
      $group_contacts = crm_get_group_contacts($group, array('display_name'));
      foreach($group_contacts as $contact) {
        $result[$contact->contact_id] = $contact->display_name;
      }
    }
    break;
    
  case 'activity':
    require_once 'CRM/Core/PseudoConstant.php';
    $result =& CRM_Core_PseudoConstant::activityType(true, 'id > 3');
    break;
    
  case 'tag':
    require_once 'CRM/Core/PseudoConstant.php';
    $result =& CRM_Core_PseudoConstant::tag();
    break;
      
  case 'location':
    require_once 'CRM/Core/PseudoConstant.php';
    $result =& CRM_Core_PseudoConstant::locationType(true);
    break;
    
    /* disabled for now
     case 'field':
     $fields = crm_get_class_properties('Individual', 'custom'); 
     foreach($fields as $field) {
     dprint_r($field);
     $result[$field[data_type] .':' . $field[id]] = 
     implode(' : ', array($field[name], $field[description]));
     }
     break;
    */
    
    /* disabled for now
     case 'search':
     require_once 'CRM/Core/PseudoConstant.php';
     $result =& CRM_Core_PseudoConstant::savedSearch();
     break; */
    
  }
  //For now, if we had to restrict the query, do not cache
  if (!$params)
    $cache[$type] = $result;

  return $result;
}



/**
 * Forms API Widget for selecting groups
 *
 * @ingroup CiviNode_Form_Widgets
 * @return array of Forms API parameters
 */
function civinode_group_selector($title, $desc = NULL, $default_gid = 0,
                                 $params = NULL){
  $groups = civinode_get_groups($params);
  //Since none is an option...
  $groups[0] = t('<select a group>');
  $widget = array(
                  '#type' => 'select',
                  '#title' => $title,
                  '#options' => $groups
                  );
  //if ($default_gid)
  $widget['#default_value'] = $default_gid;
  if ($desc)
    $widget['#description'] = $desc;

  return $widget;

}


/**
 * Forms API Widget for selecting CiviCRM profiles
 *
 * @ingroup CiviNode_Form_Widgets
 * @return array of Forms API parameters
 */
function civinode_profile_selector($title, $desc = NULL, $default_pid = 0,
                                   $params = NULL){
  $options = civinode_get_profiles($params);
  $options[0] = t('-- select a profile --  ');
  $widget = array(
                  '#type' => 'select',
                  '#title' => $title,
                  '#options' => $options
                  );
  if ($default_pid)
    $widget['#default_value'] = $default_pid;
  if ($desc)
    $widget['#description'] = $desc;

  return $widget;


}


/**
 * A forms api widget for a autocompleting contact widget
 *
 * @ingroup CiviNode_Form_Widgets
 */
function civinode_contact_auto_selector($title, $desc = NULL,
                                        $autocomplete = 'civinode/autocomplete/contact',
                                        $key='any', $val = 'any'){
  $widget = array(
                  '#type' => 'textfield',
                  '#title' => $title,
                  '#size' => 60,
                  '#autocomplete_path' => $autocomplete . "/$key/$val"
                  );
  if ($desc)
    $widget['#description'] = $desc;

  return $widget;


}


/**
 * Editing widget for a profile, wrapped for Forms API
 *
 * For now, we directly use CiviCRM UI, since we don't have quite
 * enough metadata available to us to do this right ourselves.
 *
 * If called w/o a contact_id, this widget will put up the UI for
 * creating a new contact.
 *
 * @ingroup CiviNode_Form_Widgets
 */
function civinode_get_edit_profile_html($pid, $contact_id = 0){
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), WATCHDOG_ERROR);
    return NULL;
  }

  $action = $contact_id ? 
    CIVINODE_MAGIC_ACTION_UPDATE : 
    CIVINODE_MAGIC_ACTION_ADD;

  if (!isset($_POST['op']) or $_POST['op'] == t('Delete') 
      or $_POST['op'] == t('Preview')) {
    //$action += CIVINODE_MAGIC_ACTION_PREVIEW;
    unset($_POST['_qf_default']);
    unset($_REQUEST['_qf_default']);
  }
  $title = crm_uf_get_profile_title($pid);
  $reset = isset($_POST['edit']) ? FALSE : TRUE;
  $output = crm_uf_get_profile_html( $contact_id, $title, $action, FALSE, $reset );
  //Now wrap it for the API.  It may make sense
  //to put this into a fieldset as well.
  return array('its_data' => array('#value' => $output),
               '#title' => $title,
               '#collapsible' => TRUE,
               '#type' => 'fieldset',
               '#weight' => -3);

}


/**
 * Handlers for relationships
 * @defgroup CiviCRM_Relationships
 */



/**
 * Fetch a relationship for a contact
 *
 * @ingroup CiviCRM_Relationships
 * @param int $contact_id CRM contact_id
 * @param array $rel_types names of relation types to retrieve.
 * @return
 */
function civinode_fetch_relationships($contact_id, $rel_types = array()){
  static $types;
  if (!civinode_check_init()){
    watchdog('CRM', t('CiviCRM must be installed for CiviNode'), 
             WATCHDOG_ERROR);
    return NULL;
  }

  //No contact, no relationship!
  if (!$contact_id)
    return FALSE;
  $contact = crm_get_contact(array('contact_id' => $contact_id));
  if (!$contact or civinode_check_error($contact))
    return FALSE;
  //First, we need type objects.  Cache 'em
  $type_objs = array();
  if ($rel_types and count($rel_types)) {
    if (!$types) {
      $all_types = crm_get_relationship_types();
      foreach ($all_types as $robj) {
        if (isset($robj->name_a_b)) {
          $types[$robj->name_a_b] = $robj;
          if ($robj->name_a_b != $robj->name_b_a)
            $types[$robj->name_b_a] = $robj;
        }
      }
    }
    foreach ($rel_types as $type_name) {
      //pull out the right objects.  For now, a missing
      //type is *not* an error (TO DO: best behavior?)
      if (isset($types[$type_name]))
        $type_objs[] = $types[$type_name];
    }
  }
  //Docs say this returns an array or null
  return crm_get_relationships($contact, NULL, $type_objs);

}


